/******************************************************************************
* DISCLAIMER

* This software is supplied by Renesas Electronics Corporation and is only 
* intended for use with Renesas products. No other uses are authorized.

* This software is owned by Renesas Electronics Corporation and is protected under 
* all applicable laws, including copyright laws.

* THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES 
* REGARDING THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, 
* INCLUDING BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
* PARTICULAR PURPOSE AND NON-INFRINGEMENT.  ALL SUCH WARRANTIES ARE EXPRESSLY 
* DISCLAIMED.

* TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS 
* ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES SHALL BE LIABLE 
* FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES 
* FOR ANY REASON RELATED TO THIS SOFTWARE, EVEN IF RENESAS OR ITS 
* AFFILIATES HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

* Renesas reserves the right, without notice, to make changes to this 
* software and to discontinue the availability of this software.  
* By using this software, you agree to the additional terms and 
* conditions found by accessing the following link:
* http://www.renesas.com/disclaimer
******************************************************************************/
/* Copyright (C) 2011 Renesas Electronics Corporation. All rights reserved.  */
/******************************************************************************	
* File Name    : command.c
* Version      : 1.00
* Device(s)    : RL78/G13
* Tool-Chain   : CubeSuite Version 1.5d
* H/W Platform : RL78/G13 Energy Meter Platform
* Description  : Command processing source File
******************************************************************************
* History : DD.MM.YYYY Version Description
******************************************************************************/

/******************************************************************************
Includes   <System Includes> , "Project Includes"
******************************************************************************/
/* Driver */
#include "r_cg_macrodriver.h"		/* CG Macro Driver */
#include "r_cg_wdt.h"			/* WDT Driver */


//Additional Driver
#include "r_cg_rtc.h"			/* RTC Driver */
#include "r_cg_port.h"			/* Port */

/* MW/Core, Common */

/* Application */

/* Debug Only */
#include <stdio.h>			/* Standard IO */
#include <stdarg.h>			/* Variant argument */
#include <string.h>			/* String Standard Lib */
#include <math.h>			/* Mathematic Standard Lib */
#include "cmd.h"			/* CMD Prompt Driver */
#include "command.h"
#include "compiler.h"


/******************************************************************************
Typedef definitions
******************************************************************************/
/* Command Invoker Function */
typedef uint8_t (*FUNC)(uint8_t *arg_str);

/* Command Item */
typedef struct tagCommandTable
{
	const uint8_t	*cmd_name;
	const uint8_t	*params;
	const uint8_t	*description;
	const FUNC		func;
} COMMAND_ITEM;

/******************************************************************************
Macro definitions
******************************************************************************/
/* Common macro */
#define		_BCD2DEC(x)		{x = (x & 0xF) + (x >> 4) * 10;}				/* 1 byte BCD to DEC Conversion */
#define		_DEC2BCD(x)		{x = (x % 10 ) | (((x / 10) << 4) & 0xF0);}		/* 1 byte DEC to BCD Conversion */

/* Command Status */
#define COMMAND_OK			0
#define COMMAND_ERROR		1

/* Command Invoker Statement & Table Length */
#define INVOKE_COMMAND(index, arg_str)	{													\
											if (cmd_table[index].func != NULL) 				\
											{												\
												(*cmd_table[index].func)(arg_str);			\
											}												\
										}
#define COMMAND_TABLE_LENGTH	(sizeof(cmd_table) / sizeof(COMMAND_ITEM))

/******************************************************************************
Imported global variables and functions (from other files)
******************************************************************************/


/******************************************************************************
Exported global variables and functions (to be accessed by other files)
******************************************************************************/


/******************************************************************************
Private global variables and functions
******************************************************************************/
/* Static variables */
/* Math error exception */
static uint16_t g_matherr_type = 0;

/* Sub-function (static for module) */
static uint8_t * COMMAND_GetScanOneParam(
	uint8_t *output,
	uint16_t size,
	uint8_t *input,
	uint8_t *ahead_str,
	uint8_t *break_str
);


/* Command Invoker */
static uint8_t COMMAND_InvokeUsage(uint8_t *arg_str);				/* Usage */
static uint8_t COMMAND_InvokeRtc(uint8_t *arg_str);                             /* RTC */
static uint8_t COMMAND_InvokeSetRtc(uint8_t *arg_str);                          /* Set RTC */
static uint8_t COMMAND_InvokeLED(uint8_t *arg_str);				/* Set LED */

/* Command Table */
static const COMMAND_ITEM	cmd_table[] = 
{
	/* ---------------------------------------------------------------------------------------------------------------------*/
	/* Command Name		Parameter					Description										Func Pointer */
	/* ---------------------------------------------------------------------------------------------------------------------*/
	{(const uint8_t *)"?"				,	(const uint8_t *)""						,	(const uint8_t *)"Help"								,	COMMAND_InvokeUsage		},
	{(const uint8_t *)"cls"				,	(const uint8_t *)""						,	(const uint8_t *)"Clear screen"							,	NULL				},		/* CMD module processed it internally */
	{(const uint8_t *)"rtc"             		,   	(const uint8_t *)""                                         	,   	(const uint8_t *)"Display current RTC time"                                     ,   	COMMAND_InvokeRtc               },
    	{(const uint8_t *)"setrtc"          		,  	(const uint8_t *)"dd/mm/yy hh:mm:ss ww"                     	,   	(const uint8_t *)"Set RTC time"                                                 ,   	COMMAND_InvokeSetRtc            },
	{(const uint8_t *)"setled"          		,  	(const uint8_t *)"LED x x"			          	,   	(const uint8_t *)"Set LED 1,2 1:on, 0:off"                                                ,   	COMMAND_InvokeLED            	},
};

/******************************************************************************
* Function Name: int16_t EM_atoi(uint8_t *str)
* Description  : Convert a int numeric string to a int number
* Arguments    : uint8_t *str: The address of int numeric string
* Return Value : int type,
*              : represent for the result of conversion
******************************************************************************/
int16_t EM_atoi(uint8_t *str)
{
    int16_t rlt = 0;
    int8_t sign = 1;        /* positive */  
    
    /* RL78G13 / P_003: Carelessness on checking parameter */
    /* check for NULL or empty string */
    if ((str == 0) || (*str == 0))
    {
        return 0;
    }
    
    /* trim left and check the sign */
    while (*str != '.' && *str != 0 && (*str < '1' || *str > '9'))
    {
        if (*str == '-')
        {
            sign = 0;   /* negative */
        }
        
        str++;
    }

    /* scan value, from the left */
    while (*str != 0 && *str >= '0' && *str <= '9')
    {   
        rlt *= 10;
        rlt += *str++ - '0';
    }
    
    if (sign == 0)
    {
        return (-rlt);
    }
    
    return (rlt);
}


/******************************************************************************
* Function Name    : static uint8_t *COMMAND_GetScanOneParam(
*                  :     uint8_t *output,
*                  :     uint16_t size,
*                  :     uint8_t *input,
*                  :     uint8_t *ahead_str,
*                  :     uint8_t *break_str
*                  : );
* Description      : Command Get One Parameter
* Arguments        : uint8_t *output: Output buffer
*                  : uint16_t size: Size of output buffer
*                  : uint8_t *input: Input buffer
*                  : uint8_t *ahead_str: Char array to skip the ahead the arg_str
*                  : uint8_t *break_str: Char array to break the scan processing
* Functions Called : None
* Return Value     : Next location of parameter
******************************************************************************/
static uint8_t *COMMAND_GetScanOneParam(uint8_t *output,
										uint16_t size,
										uint8_t *input,
										uint8_t *ahead_str,
										uint8_t *break_str)
{
	uint8_t i, j, break_len, ahead_len;
	
	/* Check parameter */
	if (output == NULL || size == 0 ||
		input  == NULL || break_str == NULL)
	{
		return NULL;
	}

	if (ahead_str != NULL && *ahead_str != 0)
	{
		ahead_len = (uint8_t)strlen((char *)ahead_str);
		
		/* skip all ahead char that found on arg_str */
		while (*input != 0)
		{
			for (j = 0; j < ahead_len; j++)
			{
				if (ahead_str[j] == *input)
					break;
			}
			
			/* is found in break_str? */
			if (j != ahead_len)
			{
				input++;
			}
			else	/* Not found */
			{
				break;	/* Break the outsize loop */
			}
		}
	}
	
	/* Is have remain char? */
	if (*input != 0)
	{	
		/* clear output buffer */
		memset(output, 0, size);
		
		/* Get the lenght of break_str */
		break_len = (uint8_t)strlen((char *)break_str);
	
		/* Scan the parameter */
		i = 0;
		while (i < size && *input != 0)
		{
			for (j = 0; j < break_len; j++)
			{
				if (break_str[j] == *input)
					break;
			}

			/* is found in break_str?
			 * --> end of scanning */
			if (j != break_len)
			{
				break;	/* end of scanning */
			}
			else	/* Not found --> continue to scan */
			{
				/* Update to buffer */
				output[i] = *input;
			
				/* scan next byte */
				input++;
				i++;
			}
		}
		
		return input;
	}
	else
	{
		return NULL;
	}
}

/******************************************************************************
* Function Name    : static uint8_t COMMAND_InvokeUsage(uint8_t *arg_str)
* Description      : Command Invoke Usage
* Arguments        : uint8_t index: Command index
*                  : uint8_t *arg_str: Arguments string
* Functions Called : None
* Return Value     : uint8_t, execution code, 0 is success
******************************************************************************/
static uint8_t COMMAND_InvokeUsage(uint8_t *arg_str)
{
	uint8_t i;
	
	CMD_SendString((uint8_t *)"\n\r");
	CMD_SendString((uint8_t *)"------------------------------------------------------------\n\r");
	CMD_Printf((uint8_t *)" %- 15s %- 20s %s\n\r", "Command Name", "Parameter", "Description");
	CMD_SendString((uint8_t *)"------------------------------------------------------------\n\r");
	for (i = 0; i < COMMAND_TABLE_LENGTH; i++)
	{
		CMD_Printf(	(uint8_t *)" %- 15s %- 20s %s\n\r",
					cmd_table[i].cmd_name,
					cmd_table[i].params,
					cmd_table[i].description);
	}
	
	return 0;
}

/******************************************************************************
* Function Name    : void COMMAND_PollingProcessing(void)
* Description      : COMMAND Polling Processing
* Arguments        : None
* Functions Called : None
* Return Value     : None
******************************************************************************/
void COMMAND_PollingProcessing(void)
{
	uint8_t	i, is_found_cmd;
	uint8_t *pstr, *p_strarg;
	uint8_t	cmdstr[64];

	/* Processing command line */
	if (CMD_IsCmdReceived())
	{
		/* Get command */
		pstr = CMD_Get();
		memset(cmdstr, 0, 64);

		/* Skip all ahead space */
		while (pstr[0] == ' ' && pstr[0] != 0)
		{
			pstr++;
		}
		
		/* Get command name */
		i = 0;
		while (pstr[0] != ' ' && pstr[0] != 0)
		{
			cmdstr[i++] = *pstr;
			pstr++;
		}
		
		/* Skip all ahead space on argument list */
		while (pstr[0] == ' ' && pstr[0] != 0)
		{
			pstr++;
		}
		p_strarg = pstr;

		/* is end of string */
		if (p_strarg[0] == 0)
		{
			p_strarg = NULL;	/* No parameter */
		}

		/* Find to invoke the command on the cmd_table */
		if (cmdstr[0] != 0)
		{
			is_found_cmd = 0;		/* Not found */
			for (i = 0; i < COMMAND_TABLE_LENGTH; i++)
			{
				/* is matched? */
				if (strcmp((const char *)cmdstr, (const char *)cmd_table[i].cmd_name) == 0)
				{
					INVOKE_COMMAND(i, p_strarg);		/* Invoke to function */
					is_found_cmd = 1;					/* Found command */
				}
			}
			
			/* is NOT found command? */
			if (!is_found_cmd)
			{
				CMD_SendString((uint8_t *)"\n\rUnknown command: [");
				CMD_SendString(CMD_Get());
				CMD_SendString((uint8_t *)"]\n\rType ? for command list\n\r");
			}
		}

		CMD_AckCmdReceived();
	}

	/* Math Error Exception Printout */
	if (g_matherr_type != 0)
	{
		CMD_Printf(	(uint8_t *)"\n\rDEBUG: Math error exception [type: %i]\n\r",
					g_matherr_type);

		CMD_AckCmdReceived();
		
		/* ACK */
		g_matherr_type = 0;
	}
}

/***********************************************************************************************************************
* Function Name    : static uint8_t COMMAND_LED(uint8_t *arg_str)
* Description      : Command Invoke LEDs
* Arguments        : uint8_t index: Command index
*                  : uint8_t *arg_str: Arguments string
* Functions Called : None
* Return Value     : uint8_t, execution code, 0 is success
***********************************************************************************************************************/
static uint8_t COMMAND_InvokeLED(uint8_t *arg_str)
{
    uint8_t LED;
    uint8_t buffer[20];
    
    /*Select LED 1 or 2*/
    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" ", (uint8_t *)" /");
    
    if (arg_str != NULL &&
        (buffer[0] >= '1' && buffer[0] <= '2'))
    {   
        if (buffer[0] == '1')
	{
		LED = 1;
	}
	else
	{
		LED = 2;
	}
    }
    else
    {
        CMD_SendString((uint8_t *)"Parameter error\n\r");
        return 1;
    }
    
    /*Select LED State*/
    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" ", (uint8_t *)" /");
    CMD_SendString((uint8_t *)"\n\r");
    
    if (arg_str != NULL &&
        (buffer[0] >= '0' && buffer[0] <= '1'))
    {   
        if (buffer[0] == '0')
	{
		if (LED == 1)
		{
			//LED 1 Off
			if (BIT_SELECT(P1,0)==1)
			{
				CMD_SendString((uint8_t *)"Invalid, LED 1 is OFF...\n\r");	
			}
			else
			{
				LED_1_OFF; 
				CMD_SendString((uint8_t *)"LED 1 Off\n\r");
			}
		}
		else
		{
			if (BIT_SELECT(P1,1)==1)
			{
				CMD_SendString((uint8_t *)"Invalid, LED 2 is OFF...\n\r");	
			}
			else
			{
				LED_2_OFF; //LED 2 Off
				CMD_SendString((uint8_t *)"LED 2 Off\n\r");
			}
		}
		
	}
	if (buffer[0] == '1')
	{
		if (LED == 1)
		{
			if (BIT_SELECT(P1,0)==0)
			{
				CMD_SendString((uint8_t *)"Invalid, LED 1 is ON...\n\r");	
			}
			else
			{
				LED_1_ON; //LED 1 ON
				CMD_SendString((uint8_t *)"LED 1 On\n\r");
			}
		}
		else
		{
			if (BIT_SELECT(P1,1)==0)
			{
				CMD_SendString((uint8_t *)"Invalid, LED 2 is ON...\n\r");	
			}
			else
			{
				LED_2_ON; //LED 2 On
				CMD_SendString((uint8_t *)"LED 2 On\n\r");
			}
		}
	}
    }
    else
    {
        CMD_SendString((uint8_t *)"Parameter error\n\r");
        return 1;
    }
    
    
                
    return 0;
}

/***********************************************************************************************************************
* Function Name    : static uint8_t COMMAND_InvokeRtc(uint8_t *arg_str)
* Description      : Command Invoke RTC
* Arguments        : uint8_t index: Command index
*                  : uint8_t *arg_str: Arguments string
* Functions Called : None
* Return Value     : uint8_t, execution code, 0 is success
***********************************************************************************************************************/
static uint8_t COMMAND_InvokeRtc(uint8_t *arg_str)
{
    if (LVDVRTCF == 1)
    {
    CMD_SendString((uint8_t *)"\n\r");
    CMD_SendString((uint8_t *)"Feature Unavailable!	RTC is Disabled.....\n\r");
    CMD_SendString((uint8_t *)"Please insert battery and reset to enable RTC...\n\r");
    }	
    else
    {
    rtc_counter_value_t rtctime;
    
    R_RTC_Get_CalendarCounterValue(&rtctime);
    
    CMD_Printf((uint8_t *)"\n\rRTC Time: %02x/%02x/20%02x %02x:%02x:%02x %02x\n\r",
                rtctime.day,
                rtctime.month,
                rtctime.year,
                rtctime.hour,
                rtctime.min,
                rtctime.sec,
                rtctime.week);
    }
                
    return 0;
}

/***********************************************************************************************************************
* Function Name    : static uint8_t COMMAND_InvokeSetRtc(uint8_t *arg_str)
* Description      : Command Invoke Set RTC
* Arguments        : uint8_t index: Command index
*                  : uint8_t *arg_str: Arguments string
* Functions Called : None
* Return Value     : uint8_t, execution code, 0 is success
***********************************************************************************************************************/
static uint8_t COMMAND_InvokeSetRtc(uint8_t *arg_str)
{   
    if (LVDVRTCF == 1)
    {
	    CMD_SendString((uint8_t *)"\n\r");
	    CMD_SendString((uint8_t *)"Feature Unavailable!	RTC is Disabled.....\n\r");
	    CMD_SendString((uint8_t *)"Please insert battery and reset to enable RTC...\n\r");
    }	
    else
    {
	    uint8_t status;
	    rtc_counter_value_t rtctime;
	    uint8_t buffer[20];

	    CMD_SendString((uint8_t *)"\n\rParameter(s): ");
	    CMD_SendString((uint8_t *)arg_str);
	    CMD_SendString((uint8_t *)"\n\r");
	    
	    /* Get Day parameter */
	    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" ", (uint8_t *)" /");
	    
	    if (arg_str != NULL &&
	       (buffer[0] >= '0' && buffer[0] <= '9'))
	    {   
	        /* get number */
	        rtctime.day = (uint8_t)EM_atoi(buffer);
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"Parameter error\n\r");
	        return 1;
	    }
	    
	    /* Get Month parameter */
	    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" /", (uint8_t *)" /");
	    
	    if (arg_str != NULL &&
	        (buffer[0] >= '0' && buffer[0] <= '9'))
	    {   
	        /* get number */
	        rtctime.month = (uint8_t)EM_atoi(buffer);
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"Parameter error\n\r");
	        return 1;
	    }
	    
	    /* Get Year parameter */
	    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" /", (uint8_t *)" ");
	    
	    if (arg_str != NULL &&
	        (buffer[0] >= '0' && buffer[0] <= '9'))
	    {   
	        /* get number */
	        rtctime.year = (uint8_t)EM_atoi(buffer);
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"Parameter error\n\r");
	        return 1;
	    }
	    
	    /* Get Hour parameter */
	    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" ", (uint8_t *)" :");
	    
	    if (arg_str != NULL &&
	        (buffer[0] >= '0' && buffer[0] <= '9'))
	    {   
	        /* get number */
	        rtctime.hour = (uint8_t)EM_atoi(buffer);
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"Parameter error\n\r");
	        return 1;
	    }

	    /* Get Min parameter */
	    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" :", (uint8_t *)" :");
	    
	    if (arg_str != NULL &&
	        (buffer[0] >= '0' && buffer[0] <= '9'))
	    {   
	        /* get number */
	        rtctime.min = (uint8_t)EM_atoi(buffer);
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"Parameter error\n\r");
	        return 1;
	    }
	    
	    /* Get Sec parameter */
	    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" :", (uint8_t *)" ");
	    
	    if (arg_str != NULL &&
	        (buffer[0] >= '0' && buffer[0] <= '9'))
	    {   
	        /* get number */
	        rtctime.sec = (uint8_t)EM_atoi(buffer);
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"Parameter error\n\r");
	        return 1;
	    }
	    
	    /* Get Week parameter */ 
	    arg_str = COMMAND_GetScanOneParam(buffer, 20, arg_str, (uint8_t *)" ", (uint8_t *)" ");
	    
	    if (arg_str != NULL &&
	        (buffer[0] >= '0' && buffer[0] <= '9')) /*Day of the week ( 0 = Sun to 6 = Sat)*/
	    {   
	        /* get number */
	        rtctime.week = (uint8_t)EM_atoi(buffer);
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"Parameter error\n\r");
	        return 1;
	    }
	    
	    _DEC2BCD(rtctime.day);
	    _DEC2BCD(rtctime.month);
	    _DEC2BCD(rtctime.year);
	    _DEC2BCD(rtctime.hour);
	    _DEC2BCD(rtctime.min);
	    _DEC2BCD(rtctime.sec);

	    CMD_SendString((uint8_t *)"Set RTC time...");
	    
	    CMD_SendString((uint8_t *)"\n\r");
	    
	    CMD_Printf((uint8_t *)"Set RTC time...%02X/%02X/20%02X %02X:%02X:%02X %02X\n\r",
	    rtctime.day,
	    rtctime.month,
	    rtctime.year,
	    rtctime.hour,
	    rtctime.min,
	    rtctime.sec,
	    rtctime.week
	    );
	    
	    R_RTC_Set_CalendarCounterValue(rtctime);
	    status = MD_OK;
	    
	    if (status == MD_OK)
	    {
	        CMD_SendString((uint8_t *)"OK\n\r");
	    }
	    else
	    {
	        CMD_SendString((uint8_t *)"FAILED\n\r");
	    }
	    
	    //Restart RTC after Setting
	    R_RTC_Start();
    
    }
    
    return 0;
}
